@@ -1,4 +1,4 @@
-# Copyright 2021 The Brax Authors.
+# Copyright 2022 The Brax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -19,6 +19,7 @@ import functools
 from typing import Callable, Optional, Union, overload
 
 import brax
+from brax.envs import acrobot
 from brax.envs import ant
 from brax.envs import fast
 from brax.envs import fetch
@@ -31,6 +32,7 @@ from brax.envs import inverted_double_pendulum
 from brax.envs import inverted_pendulum
 from brax.envs import reacher
 from brax.envs import reacherangle
+from brax.envs import swimmer
 from brax.envs import ur5e
 from brax.envs import walker2d
 from brax.envs import wrappers
@@ -39,6 +41,7 @@ import gym
 
 
 _envs = {
+    'acrobot': acrobot.Acrobot,
     'ant': ant.Ant,
     'fast': fast.Fast,
     'fetch': fetch.Fetch,
@@ -51,16 +54,22 @@ _envs = {
     'inverted_double_pendulum': inverted_double_pendulum.InvertedDoublePendulum,
     'reacher': reacher.Reacher,
     'reacherangle': reacherangle.ReacherAngle,
+    'swimmer': swimmer.Swimmer,
     'ur5e': ur5e.Ur5e,
     'walker2d': walker2d.Walker2d,
 }
 
 
+def get_environment(env_name, **kwargs):
+  return _envs[env_name](**kwargs)
+
+
 def create(env_name: str,
            episode_length: int = 1000,
            action_repeat: int = 1,
            auto_reset: bool = True,
            batch_size: Optional[int] = None,
+           eval_metrics: bool = False,
            **kwargs) -> Env:
   """Creates an Env with a specified brax system."""
   env = _envs[env_name](**kwargs)
@@ -70,6 +79,8 @@ def create(env_name: str,
     env = wrappers.VectorWrapper(env, batch_size)
   if auto_reset:
     env = wrappers.AutoResetWrapper(env)
+  if eval_metrics:
+    env = wrappers.EvalWrapper(env)
 
   return env  # type: ignore
 
