@@ -1,4 +1,4 @@
-# Copyright 2021 The Brax Authors.
+# Copyright 2022 The Brax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -22,8 +22,9 @@ from brax.envs import env
 class InvertedPendulum(env.Env):
   """Trains an inverted pendulum to remain stationary."""
 
-  def __init__(self, **kwargs):
-    super().__init__(_SYSTEM_CONFIG, **kwargs)
+  def __init__(self, legacy_spring=False, **kwargs):
+    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
+    super().__init__(config=config, **kwargs)
 
   def reset(self, rng: jp.ndarray) -> env.State:
     """Resets the environment to an initial state."""
@@ -59,66 +60,114 @@ class InvertedPendulum(env.Env):
     # some pre-processing to pull joint angles and velocities
     (joint_angle,), (joint_vel,) = self.sys.joints[0].angle_vel(qp)
 
-    # qpos:
-    qpos = [qp.pos[0, 2:], qp.rot[0], joint_angle]
-
-    # qvel:
-    # velocity of the cart
-    # joint angle velocities
-    qvel = [qp.vel[0], joint_vel]
-
-    return jp.concatenate(qpos + qvel)
+    # [cart pos, angle, cart vel, angle vel]
+    obs = jp.concatenate([qp.pos[0, :1], joint_angle, qp.vel[0, :1], joint_vel])
+    return obs
 
 
 _SYSTEM_CONFIG = """
-bodies {
-  name: "cart"
-  colliders {
+  bodies {
+    name: "cart"
+    colliders {
+      rotation {
+      x: 90
+      z: 90
+      }
+      capsule {
+        radius: 0.1
+        length: 0.4
+      }
+    }
+    frozen { position { x:0 y:1 z:1 } rotation { x:1 y:1 z:1 } }
+    mass: 10.471975
+  }
+  bodies {
+    name: "pole"
+    colliders {
+      capsule {
+        radius: 0.049
+        length: 0.69800085
+      }
+    }
+    frozen { position { x: 0 y: 1 z: 0 } rotation { x: 1 y: 0 z: 1 } }
+    mass: 5.0185914
+  }
+  joints {
+    name: "hinge"
+    parent: "cart"
+    child: "pole"
+    child_offset { z: -.3 }
     rotation {
-    x: 90
-    z: 90
+      z: 90.0
     }
-    capsule {
-      radius: 0.1
-      length: 0.4
+    angle_limit { min: 0.0 max: 0.0 }
+  }
+  forces {
+    name: "cart_thruster"
+    body: "cart"
+    strength: 100.0
+    thruster {}
+  }
+  collide_include {}
+  gravity {
+    z: -9.81
+  }
+  dt: 0.04
+  substeps: 8
+  dynamics_mode: "pbd"
+  """
+
+
+_SYSTEM_CONFIG_SPRING = """
+  bodies {
+    name: "cart"
+    colliders {
+      rotation {
+      x: 90
+      z: 90
+      }
+      capsule {
+        radius: 0.1
+        length: 0.4
+      }
     }
+    frozen { position { x:0 y:1 z:1 } rotation { x:1 y:1 z:1 } }
+    mass: 10.471975
   }
-  frozen { position { x:0 y:1 z:1 } rotation { x:1 y:1 z:1 } }
-  mass: 10.471975
-}
-bodies {
-  name: "pole"
-  colliders {
-    capsule {
-      radius: 0.049
-      length: 0.69800085
+  bodies {
+    name: "pole"
+    colliders {
+      capsule {
+        radius: 0.049
+        length: 0.69800085
+      }
     }
+    frozen { position { x: 0 y: 1 z: 0 } rotation { x: 1 y: 0 z: 1 } }
+    mass: 5.0185914
   }
-  frozen { position { x: 0 y: 1 z: 0 } rotation { x: 1 y: 0 z: 1 } }
-  mass: 5.0185914
-}
-joints {
-  name: "hinge"
-  stiffness: 10000.0
-  parent: "cart"
-  child: "pole"
-  child_offset { z: -.3 }
-  rotation {
-    z: 90.0
+  joints {
+    name: "hinge"
+    stiffness: 10000.0
+    parent: "cart"
+    child: "pole"
+    child_offset { z: -.3 }
+    rotation {
+      z: 90.0
+    }
+    limit_strength: 0.0
+    angle_limit { min: 0.0 max: 0.0 }
   }
-  limit_strength: 0.0
-  angle_limit { min: 0.0 max: 0.0 }
-}
-forces {
-  name: "cart_thruster"
-  body: "cart"
-  strength: 100.0
-  thruster {}
-}
-collide_include {}
-gravity {
-  z: -9.81
-}
-dt: 0.04
-substeps: 8
-"""
+  forces {
+    name: "cart_thruster"
+    body: "cart"
+    strength: 100.0
+    thruster {}
+  }
+  collide_include {}
+  gravity {
+    z: -9.81
+  }
+  dt: 0.04
+  substeps: 8
+  dynamics_mode: "legacy_spring"
+  """
