@@ -1,4 +1,4 @@
-# Copyright 2021 The Brax Authors.
+# Copyright 2022 The Brax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,10 +14,11 @@
 
 """Wrappers for Brax and Gym env."""
 
-from typing import ClassVar, Optional
+from typing import ClassVar, Dict, Optional
 
 from brax import jumpy as jp
 from brax.envs import env as brax_env
+import flax
 import gym
 from gym import spaces
 from gym.vector import utils
@@ -25,7 +26,7 @@ import jax
 
 
 class VectorWrapper(brax_env.Wrapper):
-  """Vectorizes Brax env."""
+  """DEPRECATED Vectorizes Brax env. Use VmapWrapper instead."""
 
   def __init__(self, env: brax_env.Env, batch_size: int):
     super().__init__(env)
@@ -39,6 +40,16 @@ class VectorWrapper(brax_env.Wrapper):
     return jp.vmap(self.env.step)(state, action)
 
 
+class VmapWrapper(brax_env.Wrapper):
+  """Vectorizes Brax env."""
+
+  def reset(self, rng: jp.ndarray) -> brax_env.State:
+    return jp.vmap(self.env.reset)(rng)
+
+  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
+    return jp.vmap(self.env.step)(state, action)
+
+
 class EpisodeWrapper(brax_env.Wrapper):
   """Maintains episode step count and sets done at episode end."""
 
@@ -97,6 +108,62 @@ class AutoResetWrapper(brax_env.Wrapper):
     return state.replace(qp=qp, obs=obs)
 
 
+@flax.struct.dataclass
+class EvalMetrics:
+  current_episode_metrics: Dict[str, jp.ndarray]
+  completed_episodes_metrics: Dict[str, jp.ndarray]
+  completed_episodes: jp.ndarray
+  completed_episodes_steps: jp.ndarray
+
+
+class EvalWrapper(brax_env.Wrapper):
+  """Brax env with eval metrics."""
+
+  def reset(self, rng: jp.ndarray) -> brax_env.State:
+    reset_state = self.env.reset(rng)
+    reset_state.metrics['reward'] = reset_state.reward
+    eval_metrics = EvalMetrics(
+        current_episode_metrics=jax.tree_map(jp.zeros_like,
+                                             reset_state.metrics),
+        completed_episodes_metrics=jax.tree_map(
+            lambda x: jp.zeros_like(jp.sum(x)), reset_state.metrics),
+        completed_episodes=jp.zeros(()),
+        completed_episodes_steps=jp.zeros(()))
+    reset_state.info['eval_metrics'] = eval_metrics
+    return reset_state
+
+  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
+    state_metrics = state.info['eval_metrics']
+    if not isinstance(state_metrics, EvalMetrics):
+      raise ValueError(
+          f'Incorrect type for state_metrics: {type(state_metrics)}')
+    del state.info['eval_metrics']
+    nstate = self.env.step(state, action)
+    nstate.metrics['reward'] = nstate.reward
+    # steps stores the highest step reached when done = True, and then
+    # the next steps becomes action_repeat
+    completed_episodes_steps = state_metrics.completed_episodes_steps + jp.sum(
+        nstate.info['steps'] * nstate.done)
+    current_episode_metrics = jax.tree_multimap(
+        lambda a, b: a + b, state_metrics.current_episode_metrics,
+        nstate.metrics)
+    completed_episodes = state_metrics.completed_episodes + jp.sum(nstate.done)
+    completed_episodes_metrics = jax.tree_multimap(
+        lambda a, b: a + jp.sum(b * nstate.done),
+        state_metrics.completed_episodes_metrics, current_episode_metrics)
+    current_episode_metrics = jax.tree_multimap(
+        lambda a, b: a * (1 - nstate.done) + b * nstate.done,
+        current_episode_metrics, nstate.metrics)
+
+    eval_metrics = EvalMetrics(
+        current_episode_metrics=current_episode_metrics,
+        completed_episodes_metrics=completed_episodes_metrics,
+        completed_episodes=completed_episodes,
+        completed_episodes_steps=completed_episodes_steps)
+    nstate.info['eval_metrics'] = eval_metrics
+    return nstate
+
+
 class GymWrapper(gym.Env):
   """A wrapper that converts Brax Env to one that follows Gym API."""
 
@@ -132,16 +199,18 @@ class GymWrapper(gym.Env):
 
     def step(state, action):
       state = self._env.step(state, action)
-      return state, state.obs, state.reward, state.done, state.info
+      return state, state.obs, state.reward, state.done, state.metrics
 
     self._step = jax.jit(step, backend=self.backend)
 
   def reset(self):
     self._state, obs, self._key = self._reset(self._key)
+    # We return device arrays for pytorch users.
     return obs
 
   def step(self, action):
     self._state, obs, reward, done, info = self._step(self._state, action)
+    # We return device arrays for pytorch users.
     return obs, reward, done, info
 
   def seed(self, seed: int = 0):
@@ -202,7 +271,7 @@ class VectorGymWrapper(gym.vector.VectorEnv):
 
     def step(state, action):
       state = self._env.step(state, action)
-      return state, state.obs, state.reward, state.done, state.info
+      return state, state.obs, state.reward, state.done, state.metrics
 
     self._step = jax.jit(step, backend=self.backend)
 
