@@ -1,4 +1,4 @@
-# Copyright 2021 The Brax Authors.
+# Copyright 2022 The Brax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -37,8 +37,7 @@ class Body:
 
   def __init__(self, config: config_pb2.Config):
     self.idx = jp.arange(0, len(config.bodies))
-    inertias = [jp.inv(jp.diag(vec_to_arr(b.inertia))) for b in config.bodies]
-    self.inertia = jp.array(inertias)
+    self.inertia = 1. / jp.array([vec_to_arr(b.inertia) for b in config.bodies])
     self.mass = jp.array([b.mass for b in config.bodies])
     self.active = jp.array(
         [0.0 if b.frozen.all else 1.0 for b in config.bodies])
@@ -56,7 +55,7 @@ class Body:
       dP: An impulse to apply to this body
     """
     dvel = impulse / self.mass
-    dang = jp.matmul(self.inertia, jp.cross(pos - qp.pos, impulse))
+    dang = self.inertia * jp.cross(pos - qp.pos, impulse)
     return P(vel=dvel, ang=dang)
 
 
