@@ -1,4 +1,4 @@
-# Copyright 2021 The Brax Authors.
+# Copyright 2022 The Brax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,6 +17,7 @@
 from absl.testing import absltest
 from brax import envs
 from brax import jumpy as jp
+import jax
 
 
 class GraspTest(absltest.TestCase):
@@ -24,15 +25,30 @@ class GraspTest(absltest.TestCase):
   def testGrasp(self):
     env = envs.create('grasp')
     grasp_action = jp.array([
-        -.4, -.35, -1, -0,  # gripper arm 1
-        .4, .35, 1, 0,      # gripper arm 2
-        .4, .35, 1, 0,      # gripper arm 3
-        .4, .35, 1, 0,      # gripper arm 4
+        -.1, 0., 0., -.35, 0., 0., -1, -0, 0.,  # gripper arm 1
+          .1, 0., 0.,  .35, 0., 0.,  1,  0, 0.,     # gripper arm 2
+          .1, 0., 0.,  .35, 0., 0.,  1,  0, 0.,      # gripper arm 3
+          .1, 0., 0.,  .35, 0., 0.,  1,  0, 0.,      # gripper arm 4
         0, 0, -.9           # position action
     ])
+
+    def raise_action(i):
+      return jp.array([
+        -.1, 0., 0., -.35, 0., 0., -1, -0, 0.,  # gripper arm 1
+          .1, 0., 0.,  .35, 0., 0.,  1,  0, 0.,     # gripper arm 2
+          .1, 0., 0.,  .35, 0., 0.,  1,  0, 0.,      # gripper arm 3
+          .1, 0., 0.,  .35, 0., 0.,  1,  0, 0.,      # gripper arm 4
+        0, 0, -.9*((250 - i)/250) + -.5*(i/250)           # position action
+      ])
+
     state = env.reset(jp.random_prngkey(0))
-    for _ in range(500):
-      state = env.step(state, grasp_action)
+    step = jax.jit(env.step)
+    # grasp
+    for _ in range(250):
+      state = step(state, grasp_action)
+    # slowly lift
+    for i in range(250):
+      state = step(state, raise_action(i))
 
     self.assertGreater(state.qp.pos[1, 2], 1.36)  # ball lifted off ground
     self.assertLess(state.qp.ang[1, 2], .01)  # ball not rolling
